/*
  Copyright 1996, 2002 Ronald S. Burkey.

  This file is part of Ron's Indexing Program (RIP).

  RIP is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  RIP is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with RIP; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:	longsrt4.c 
  Target:	MS-DOS
  Compiler:	Borland Turbo C 2.0
  Purpose:	Sorts a big,big file of N-byte records.
  Mods:		10/10/96 RSB    Wrote (longsort.c/longsrt2.c).
  		10/31/96 RSB    Improved form (longsrt3.c), I hope, of 
				longsrt2.c.
  		11/03/96 RSB	Completely rewritten version (longsrt4.c) of 
				longsrt3.c.  The difference is that the sorting 
				is not done in place.  Temporary files are 
				opened, closed, and deleted, as needed, but in 
				such a way that there are never too many open 
				at once.  I hope that this will approximately 
				double the speed of the thing (for large files), 
				since at each recursion level except the lowest, 
				there had been a copy-back step that took nearly 
				as long as the merge step that occured at that 
				level.
		01/08/02 RSB	Reformatted in the GNU style, and otherwise
				cleaned up a little, in preparation for
				first GPL'd release.
				
  The basic technique, used in longsort.c through longsrt3.c, is a simple 
  recursion.  If the file is short enough to sort, does so.  Otherwise, splits 
  in half and sorts/merges the halves.  The design of the code is really
  geared toward getting around the very limited amount of memory -- and the
  segmented nature of that memory -- available in MS-DOS.  The code should
  still work -- if it can be made to compile -- on less-crippled memory
  architectures.                       	
*/

#include <stdlib.h>
#include <stdio.h>
#include <alloc.h>

#define MAX_MERGE 8
typedef struct
{
  FILE *f;
  char *buf;
  void *rec;
}
MERGE;
MERGE merge[MAX_MERGE];

typedef struct
{
  unsigned long len;
  unsigned ptr;
  unsigned long start, end;
}
DUMMIES;

unsigned inmem_level = 0;
char *mergebuffer = NULL;

/* Compares two records, assumed to be char[recsize] in MSB to LSB order. */
unsigned orecsize;
int
rec_cmp (void *r1, void *r2)
{
  char *s1, *s2;
  int i;
  s1 = orecsize + ((char *) r1);
  s2 = orecsize + ((char *) r2);
  for (i = orecsize; i >= 0; i--, s1--, s2--)
    if (*s1 < *s2)
      return (-1);
    else if (*s1 > *s2)
      return (1);
  return (0);
}

/* Returns 0 on success */
int
longsort4 (int (*rec_cmp) (void *, void *), 
	   char *infile,	/* Name of the input file */
	   unsigned recsize,	/* Size of the records */
	   unsigned long start,	/* Starting input record */
	   unsigned long end,	/* Ending input record */
	   char *outfile	/* Name of the output file */
  )
{

  FILE *fin, *fout;
  int i, retval;
  unsigned long heapleft, memwanted, bufsize;
  unsigned u, maxbufsize, memblocks;
  unsigned long vbufsize;
  DUMMIES d[MAX_MERGE];
  char mergename[20];

  void *bv, *bd;
  int whichi;

  if (end <= start)
    return (0);
  if (!inmem_level)
    {
      if (NULL == mergebuffer)
	{
	  mergebuffer = farcalloc (MAX_MERGE, recsize);
	  if (mergebuffer == NULL)
	    return (13);
	}
    }
  inmem_level++;
  orecsize = recsize - 1;
  fout = NULL;
  for (i = 0; i < MAX_MERGE; i++)
    {
      merge[i].f = NULL;
      merge[i].buf = NULL;
    }

  /* Here we compare the available RAM against the desired sort size.  If
     there is any way we can fit the entire thing in RAM while still
     obeying the MAX_MERGE constraint, we will.  If not, then we will split
     the file and merge the chunks.  However, if the entire thing WILL fit
     into memory, we still need to divide the in-memory stuff into blocks
     of 64K or less to meet dumb 8086 restrictions.  We do this by choosing
     the minimum number of blocks we can get away with, and then dividing
     the file evenly across them. */
  maxbufsize = 0xff00U / recsize;
  memwanted = end - start;
  heapleft = (farcoreleft () - 1024) / recsize;
#pragma warn -sig
  memblocks = (memwanted + maxbufsize - 1) / maxbufsize;
#pragma warn +sig
  if (memblocks > MAX_MERGE)
    goto split_and_merge;
  bufsize = (memwanted + memblocks - 1) / memblocks;
  if (bufsize > maxbufsize)
    goto split_and_merge;
  if ((bufsize * 1L * memblocks) > heapleft)
    goto split_and_merge;

  /* Do entirely in memory **************************************** */
  fin = fopen (infile, "rb");
  if (fin == NULL)
    {
      retval = 1;
      goto done;
    }
  if (fseek (fin, end * recsize, SEEK_SET))
    {
      retval = 2;
      goto done;
    }
  if (fseek (fin, start * recsize, SEEK_SET))
    {
      retval = 3;
      goto done;
    }

  /* The case of the entire thing fitting into a single memory block is
     special, just because it's faster. */
  if (memblocks == 1)
    {

      merge[0].buf = farcalloc (bufsize, recsize);
      if (merge[0].buf == NULL)
	{
	  retval = 4;
	  goto done;
	}
#pragma warn -sig
      if (bufsize != fread (merge[0].buf, recsize, bufsize, fin))
	{
	  retval = 7;
	  goto done;
	}
      qsort (merge[0].buf, bufsize, recsize, rec_cmp);
      fout = fopen (outfile, "wb");
      if (fout == NULL)
	{
	  retval = 5;
	  goto done;
	}
      if (bufsize != fwrite (merge[0].buf, recsize, bufsize, fout))
	{
	  retval = 6;
	  goto done;
	}
#pragma warn +sig
      fclose (fout);

      /* Otherwise, sort the pieces in memory and merge to disk. */
    }
  else
    {

      /* Set up the variables. */
#pragma warn -sig
      for (i = 1; i < memblocks; i++)
	d[i].len = bufsize;
      d[0].len = memwanted - (memblocks - 1) * bufsize;
#pragma warn +sig
      d[0].start = start;
      d[0].end = start + d[0].len;
      for (i = 1; i < memblocks; i++)
	{
	  d[i].start = d[i - 1].end;
	  d[i].end = d[i].start + d[i].len;
	}
      /*
         for (i=0; i<memblocks; i++)
         printf("%d: L=%lu S=%lu E=%lu\n",
         i,d[i].len,d[i].start,d[i].end);
       */

      /* Load chunks into memory and sort them. */
      for (i = 0; i < memblocks; i++)
	{
	  merge[i].buf = farcalloc (d[i].len, recsize);
	  if (merge[i].buf == NULL)
	    {
	      retval = 8;
	      goto done;
	    }
#pragma warn -sig
	  u = d[i].len;
#pragma warn +sig
	  if (d[i].len != fread (merge[i].buf, recsize, u, fin))
	    {
	      retval = 9;
	      goto done;
	    }
	  qsort (merge[i].buf, u, recsize, rec_cmp);
	}

      /* Merge to disk. */
      fout = fopen (outfile, "wb");
      if (fout == NULL)
	{
	  retval = 5;
	  goto done;
	}
      for (i = 0; i < memblocks; i++)
	d[i].ptr = 0;
      for (;;)
	{
	  bv = NULL;
	  for (i = 0; i < memblocks; i++)
	    if (d[i].ptr < d[i].len)
	      {
		bd = merge[i].buf + d[i].ptr * recsize;
		if (bv == NULL || rec_cmp (bd, bv) < 0)
		  {
		    bv = bd;
		    whichi = i;
		  }
	      }
	  if (bv == NULL)
	    break;
	  if (1 != fwrite (bv, recsize, 1, fout))
	    {
	      retval = 6;
	      goto done;
	    }
	  d[whichi].ptr++;
	}

    }

  retval = 0;
  goto done;

  /* Split into chunks, sort chunks, then merge ******************* */
split_and_merge:

  /* Set up the variables. */
  memblocks = MAX_MERGE;
  bufsize = (memwanted + memblocks - 1) / memblocks;
#pragma warn -sig
  for (i = 1; i < memblocks; i++)
    d[i].len = bufsize;
  d[0].len = memwanted - (memblocks - 1) * bufsize;
#pragma warn +sig
  d[0].start = start;
  d[0].end = start + d[0].len;
  for (i = 1; i < memblocks; i++)
    {
      d[i].start = d[i - 1].end;
      d[i].end = d[i].start + d[i].len;
    }

  /* Sort the chunks. Create temporary files with names derived from
     the level and the block number. */
  for (i = 0; i < memblocks; i++)
    {
      sprintf (mergename, "%u_%u.tmp", inmem_level, i);
      retval =
	longsort4 (rec_cmp, infile, recsize, d[i].start, d[i].end, mergename);
      if (retval)
	goto done;
    }

  /* Merge to disk. */
#pragma warn -sig
  vbufsize = (farcoreleft () - 2048) / (MAX_MERGE + 1);
#pragma warn +sig
  vbufsize &= 0xfffffc00L;
  if (vbufsize > 16384)
    vbufsize = 16384;
  fout = fopen (outfile, "wb");
  if (fout == NULL)
    {
      retval = 5;
      goto done;
    }
#pragma warn -sig
  if (setvbuf (fout, NULL, _IOFBF, vbufsize))
    {
      retval = 15;
      goto done;
    }
#pragma warn +sig
  for (i = 0; i < memblocks; i++)
    {
      sprintf (mergename, "%u_%u.tmp", inmem_level, i);
      if (merge[i].f != NULL)
	{
	  retval = 16;
	  goto done;
	}
      merge[i].f = fopen (mergename, "rb");
      if (merge[i].f == NULL)
	{
	  retval = 11;
	  goto done;
	}
#pragma warn -sig
      if (setvbuf (merge[i].f, NULL, _IOFBF, vbufsize))
	{
	  retval = 15;
	  goto done;
	}
#pragma warn +sig
      merge[i].rec = &mergebuffer[i * recsize];
      if (1 != fread (merge[i].rec, recsize, 1, merge[i].f))
	merge[i].rec = NULL;
    }
  for (;;)
    {
      bv = NULL;
      for (i = 0; i < memblocks; i++)
	{
	  bd = merge[i].rec;
	  if (bd != NULL)
	    {
	      if (bv == NULL || rec_cmp (bd, bv) < 0)
		{
		  bv = bd;
		  whichi = i;
		}
	    }
	}
      if (bv == NULL)
	break;
      if (1 != fwrite (bv, recsize, 1, fout))
	{
	  retval = 6;
	  goto done;
	}
      if (1 != fread (bv, recsize, 1, merge[whichi].f))
	{
	  fclose (merge[whichi].f);
	  merge[whichi].f = NULL;
	  sprintf (mergename, "%u_%u.tmp", inmem_level, whichi);
	  remove (mergename);
	  merge[whichi].rec = NULL;
	}
    }

  retval = 0;
  goto done;

  /* Done ********************************************************* */
done:
  if (NULL != fin)
    fclose (fin);
  if (NULL != fout)
    fclose (fout);
  for (i = 0; i < MAX_MERGE; i++)
    {
      if (merge[i].f != NULL)
	{
	  fclose (merge[i].f);
	  merge[i].f = NULL;
	}
      if (merge[i].buf != NULL)
	{
	  farfree (merge[i].buf);
	  merge[i].buf = NULL;
	}
    }
  inmem_level--;
  if (!inmem_level)
    {
      if (mergebuffer != NULL)
	{
	  farfree (mergebuffer);
	  mergebuffer = NULL;
	}
    }
  return (retval);
}


/*--------------------------------------------------------------------*/
/* The following is a main program that can be compiled for testing
   purposes.  In general, though, the purpose of this file is really
   just to provide library functions. */

#ifdef TESTMAIN_LONGSRT4

/* Increase the stack space allocated by Turbo C. */
extern unsigned _stklen = 0xff00U;

void
main (int argc, char *argv[])
{
  FILE *f;
  unsigned long n;
  int i;

  if (argc < 3)
    {
      printf ("USAGE:\n");
      printf ("\tLONGSORT infile outfile\n");
      exit (1);
    }
  f = fopen (argv[1], "rb");
  fseek (f, 0L, SEEK_END);
  n = ftell (f) / 6;
  fclose (f);
  i = longsort4 (rec_cmp, argv[1], 6, 0, n, argv[2]);
  printf ("Error code=%d\n", i);
  exit (0);
}

#endif

